"""Products API endpoints."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pytentacle.models.products import Product
from pytentacle.models.tariffs import HistoricalCharge

if TYPE_CHECKING:
    import builtins
    from datetime import datetime

    from pytentacle.core.base import BaseClient


class ProductsAPI:
    """Products API endpoints.

    Provides access to energy products and their tariff rates.
    """

    def __init__(self, client: BaseClient) -> None:
        """Initialize the products API.

        Args:
            client: The HTTP client to use
        """
        self.client = client

    def list(  # noqa: PLR0913
        self,
        *,
        is_variable: bool | None = None,
        is_green: bool | None = None,
        is_tracker: bool | None = None,
        is_prepay: bool | None = None,
        is_business: bool | None = None,
        available_at: datetime | None = None,
    ) -> list[Product]:
        """List all available products.

        Args:
            is_variable: Filter by variable rate products
            is_green: Filter by green/renewable products
            is_tracker: Filter by tracker products
            is_prepay: Filter by prepayment products
            is_business: Filter by business products
            available_at: Filter by products available at this date

        Returns:
            List of Product objects
        """
        params = {}
        if is_variable is not None:
            params["is_variable"] = str(is_variable).lower()
        if is_green is not None:
            params["is_green"] = str(is_green).lower()
        if is_tracker is not None:
            params["is_tracker"] = str(is_tracker).lower()
        if is_prepay is not None:
            params["is_prepay"] = str(is_prepay).lower()
        if is_business is not None:
            params["is_business"] = str(is_business).lower()
        if available_at is not None:
            params["available_at"] = available_at.isoformat()

        results = self.client.get_paginated("products/", params)
        return [Product.model_validate(item) for item in results]

    def get(self, product_code: str) -> Product:
        """Get detailed information about a specific product.

        Args:
            product_code: The product code (e.g., 'AGILE-24-04-03')

        Returns:
            Product object with all tariff details
        """
        data = self.client.get(f"products/{product_code}/")
        return Product.model_validate(data)

    def get_electricity_standard_unit_rates(
        self,
        product_code: str,
        tariff_code: str,
        *,
        period_from: datetime | None = None,
        period_to: datetime | None = None,
        page_size: int | None = None,
    ) -> builtins.list[HistoricalCharge]:
        """Get standard electricity unit rates for a tariff.

        Args:
            product_code: The product code
            tariff_code: The tariff code (e.g., 'E-1R-AGILE-24-04-03-A')
            period_from: Filter rates from this datetime (inclusive)
            period_to: Filter rates to this datetime (exclusive)
            page_size: Number of results per page (max 1500)

        Returns:
            List of HistoricalCharge objects
        """
        params: dict[str, str | int] = {}
        if period_from:
            params["period_from"] = period_from.isoformat()
        if period_to:
            params["period_to"] = period_to.isoformat()
        if page_size:
            params["page_size"] = page_size

        endpoint = (
            f"products/{product_code}/electricity-tariffs/{tariff_code}/"
            "standard-unit-rates/"
        )
        results = self.client.get_paginated(endpoint, params)
        return [HistoricalCharge.model_validate(item) for item in results]

    def get_electricity_day_unit_rates(
        self,
        product_code: str,
        tariff_code: str,
        *,
        period_from: datetime | None = None,
        period_to: datetime | None = None,
        page_size: int | None = None,
    ) -> builtins.list[HistoricalCharge]:
        """Get day unit rates for Economy 7 electricity tariff.

        Args:
            product_code: The product code
            tariff_code: The tariff code
            period_from: Filter rates from this datetime (inclusive)
            period_to: Filter rates to this datetime (exclusive)
            page_size: Number of results per page (max 1500)

        Returns:
            List of HistoricalCharge objects
        """
        params: dict[str, str | int] = {}
        if period_from:
            params["period_from"] = period_from.isoformat()
        if period_to:
            params["period_to"] = period_to.isoformat()
        if page_size:
            params["page_size"] = page_size

        endpoint = (
            f"products/{product_code}/electricity-tariffs/{tariff_code}/day-unit-rates/"
        )
        results = self.client.get_paginated(endpoint, params)
        return [HistoricalCharge.model_validate(item) for item in results]

    def get_electricity_night_unit_rates(
        self,
        product_code: str,
        tariff_code: str,
        *,
        period_from: datetime | None = None,
        period_to: datetime | None = None,
        page_size: int | None = None,
    ) -> builtins.list[HistoricalCharge]:
        """Get night unit rates for Economy 7 electricity tariff.

        Args:
            product_code: The product code
            tariff_code: The tariff code
            period_from: Filter rates from this datetime (inclusive)
            period_to: Filter rates to this datetime (exclusive)
            page_size: Number of results per page (max 1500)

        Returns:
            List of HistoricalCharge objects
        """
        params: dict[str, str | int] = {}
        if period_from:
            params["period_from"] = period_from.isoformat()
        if period_to:
            params["period_to"] = period_to.isoformat()
        if page_size:
            params["page_size"] = page_size

        endpoint = (
            f"products/{product_code}/electricity-tariffs/{tariff_code}/"
            "night-unit-rates/"
        )
        results = self.client.get_paginated(endpoint, params)
        return [HistoricalCharge.model_validate(item) for item in results]

    def get_electricity_standing_charges(
        self,
        product_code: str,
        tariff_code: str,
        *,
        period_from: datetime | None = None,
        period_to: datetime | None = None,
        page_size: int | None = None,
    ) -> builtins.list[HistoricalCharge]:
        """Get electricity standing charges for a tariff.

        Args:
            product_code: The product code
            tariff_code: The tariff code
            period_from: Filter charges from this datetime (inclusive)
            period_to: Filter charges to this datetime (exclusive)
            page_size: Number of results per page (max 1500)

        Returns:
            List of HistoricalCharge objects
        """
        params: dict[str, str | int] = {}
        if period_from:
            params["period_from"] = period_from.isoformat()
        if period_to:
            params["period_to"] = period_to.isoformat()
        if page_size:
            params["page_size"] = page_size

        endpoint = (
            f"products/{product_code}/electricity-tariffs/{tariff_code}/"
            "standing-charges/"
        )
        results = self.client.get_paginated(endpoint, params)
        return [HistoricalCharge.model_validate(item) for item in results]

    def get_gas_standard_unit_rates(
        self,
        product_code: str,
        tariff_code: str,
        *,
        period_from: datetime | None = None,
        period_to: datetime | None = None,
        page_size: int | None = None,
    ) -> builtins.list[HistoricalCharge]:
        """Get gas unit rates for a tariff.

        Args:
            product_code: The product code
            tariff_code: The tariff code (e.g., 'G-1R-AGILE-24-04-03-A')
            period_from: Filter rates from this datetime (inclusive)
            period_to: Filter rates to this datetime (exclusive)
            page_size: Number of results per page (max 1500)

        Returns:
            List of HistoricalCharge objects
        """
        params: dict[str, str | int] = {}
        if period_from:
            params["period_from"] = period_from.isoformat()
        if period_to:
            params["period_to"] = period_to.isoformat()
        if page_size:
            params["page_size"] = page_size

        endpoint = (
            f"products/{product_code}/gas-tariffs/{tariff_code}/standard-unit-rates/"
        )
        results = self.client.get_paginated(endpoint, params)
        return [HistoricalCharge.model_validate(item) for item in results]

    def get_gas_standing_charges(
        self,
        product_code: str,
        tariff_code: str,
        *,
        period_from: datetime | None = None,
        period_to: datetime | None = None,
        page_size: int | None = None,
    ) -> builtins.list[HistoricalCharge]:
        """Get gas standing charges for a tariff.

        Args:
            product_code: The product code
            tariff_code: The tariff code
            period_from: Filter charges from this datetime (inclusive)
            period_to: Filter charges to this datetime (exclusive)
            page_size: Number of results per page (max 1500)

        Returns:
            List of HistoricalCharge objects
        """
        params: dict[str, str | int] = {}
        if period_from:
            params["period_from"] = period_from.isoformat()
        if period_to:
            params["period_to"] = period_to.isoformat()
        if page_size:
            params["page_size"] = page_size

        endpoint = (
            f"products/{product_code}/gas-tariffs/{tariff_code}/standing-charges/"
        )
        results = self.client.get_paginated(endpoint, params)
        return [HistoricalCharge.model_validate(item) for item in results]



"""Tests for Products API endpoints."""

from datetime import datetime, timezone
from unittest.mock import MagicMock

import pytest

from pytentacle.api.products import ProductsAPI
from pytentacle.models.products import Product
from pytentacle.models.tariffs import HistoricalCharge


@pytest.fixture
def products_api() -> ProductsAPI:
    """Return a products API instance with mocked client."""
    client = MagicMock()
    return ProductsAPI(client)


@pytest.fixture
def mock_product_data() -> dict:
    """Return mock product data."""
    return {
        "code": "AGILE-24-04-03",
        "direction": "IMPORT",
        "full_name": "Agile Octopus April 2024 v1",
        "display_name": "Agile Octopus",
        "description": "Agile Octopus gives you half-hourly tariffs.",
        "is_variable": True,
        "is_green": True,
        "is_tracker": False,
        "is_prepay": False,
        "is_business": False,
        "is_restricted": False,
        "term": 12,
        "available_from": "2024-04-03T00:00:00Z",
        "available_to": None,
        "links": [],
        "brand": "OCTOPUS_ENERGY",
    }


@pytest.fixture
def mock_charge_data() -> dict:
    """Return mock historical charge data."""
    return {
        "value_exc_vat": 15.5,
        "value_inc_vat": 16.275,
        "valid_from": "2024-01-01T00:00:00Z",
        "valid_to": "2024-01-01T00:30:00Z",
        "payment_method": "DIRECT_DEBIT",
    }


class TestProductsList:
    """Tests for products list method."""

    def test_list_no_filters(
        self, products_api: ProductsAPI, mock_product_data: dict
    ) -> None:
        """Test listing products without filters."""
        products_api.client.get_paginated.return_value = [mock_product_data]  # type: ignore[attr-defined]

        result = products_api.list()

        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "products/", {}
        )
        assert len(result) == 1
        assert isinstance(result[0], Product)
        assert result[0].code == "AGILE-24-04-03"

    def test_list_with_is_variable(
        self, products_api: ProductsAPI, mock_product_data: dict
    ) -> None:
        """Test listing products filtered by is_variable."""
        products_api.client.get_paginated.return_value = [mock_product_data]  # type: ignore[attr-defined]

        result = products_api.list(is_variable=True)

        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "products/", {"is_variable": "true"}
        )
        assert len(result) == 1

    def test_list_with_is_green(
        self, products_api: ProductsAPI, mock_product_data: dict
    ) -> None:
        """Test listing products filtered by is_green."""
        products_api.client.get_paginated.return_value = [mock_product_data]  # type: ignore[attr-defined]

        result = products_api.list(is_green=True)

        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "products/", {"is_green": "true"}
        )
        assert len(result) == 1

    def test_list_with_is_tracker(
        self, products_api: ProductsAPI, mock_product_data: dict
    ) -> None:
        """Test listing products filtered by is_tracker."""
        products_api.client.get_paginated.return_value = [mock_product_data]  # type: ignore[attr-defined]

        result = products_api.list(is_tracker=False)

        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "products/", {"is_tracker": "false"}
        )
        assert len(result) == 1

    def test_list_with_is_prepay(
        self, products_api: ProductsAPI, mock_product_data: dict
    ) -> None:
        """Test listing products filtered by is_prepay."""
        products_api.client.get_paginated.return_value = [mock_product_data]  # type: ignore[attr-defined]

        result = products_api.list(is_prepay=True)

        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "products/", {"is_prepay": "true"}
        )
        assert len(result) == 1

    def test_list_with_is_business(
        self, products_api: ProductsAPI, mock_product_data: dict
    ) -> None:
        """Test listing products filtered by is_business."""
        products_api.client.get_paginated.return_value = [mock_product_data]  # type: ignore[attr-defined]

        result = products_api.list(is_business=False)

        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "products/", {"is_business": "false"}
        )
        assert len(result) == 1

    def test_list_with_available_at(
        self, products_api: ProductsAPI, mock_product_data: dict
    ) -> None:
        """Test listing products filtered by available_at."""
        products_api.client.get_paginated.return_value = [mock_product_data]  # type: ignore[attr-defined]
        available_at = datetime(2024, 4, 3, tzinfo=timezone.utc)

        result = products_api.list(available_at=available_at)

        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "products/", {"available_at": "2024-04-03T00:00:00+00:00"}
        )
        assert len(result) == 1

    def test_list_with_all_filters(
        self, products_api: ProductsAPI, mock_product_data: dict
    ) -> None:
        """Test listing products with all filters."""
        products_api.client.get_paginated.return_value = [mock_product_data]  # type: ignore[attr-defined]
        available_at = datetime(2024, 4, 3, tzinfo=timezone.utc)

        result = products_api.list(
            is_variable=True,
            is_green=True,
            is_tracker=False,
            is_prepay=False,
            is_business=False,
            available_at=available_at,
        )

        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "products/",
            {
                "is_variable": "true",
                "is_green": "true",
                "is_tracker": "false",
                "is_prepay": "false",
                "is_business": "false",
                "available_at": "2024-04-03T00:00:00+00:00",
            },
        )
        assert len(result) == 1


class TestProductsGet:
    """Tests for products get method."""

    def test_get(self, products_api: ProductsAPI, mock_product_data: dict) -> None:
        """Test getting a specific product."""
        products_api.client.get.return_value = mock_product_data  # type: ignore[attr-defined]

        result = products_api.get("AGILE-24-04-03")

        products_api.client.get.assert_called_once_with("products/AGILE-24-04-03/")  # type: ignore[attr-defined]
        assert isinstance(result, Product)
        assert result.code == "AGILE-24-04-03"


class TestElectricityStandardUnitRates:
    """Tests for electricity standard unit rates method."""

    def test_get_electricity_standard_unit_rates_no_params(
        self, products_api: ProductsAPI, mock_charge_data: dict
    ) -> None:
        """Test getting electricity standard unit rates without optional params."""
        products_api.client.get_paginated.return_value = [mock_charge_data]  # type: ignore[attr-defined]

        result = products_api.get_electricity_standard_unit_rates(
            "AGILE-24-04-03", "E-1R-AGILE-24-04-03-A"
        )

        expected_endpoint = (
            "products/AGILE-24-04-03/electricity-tariffs/E-1R-AGILE-24-04-03-A/"
            "standard-unit-rates/"
        )
        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {}
        )
        assert len(result) == 1
        assert isinstance(result[0], HistoricalCharge)

    def test_get_electricity_standard_unit_rates_with_dates(
        self, products_api: ProductsAPI, mock_charge_data: dict
    ) -> None:
        """Test getting electricity standard unit rates with date filters."""
        products_api.client.get_paginated.return_value = [mock_charge_data]  # type: ignore[attr-defined]
        period_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
        period_to = datetime(2024, 1, 31, tzinfo=timezone.utc)

        result = products_api.get_electricity_standard_unit_rates(
            "AGILE-24-04-03",
            "E-1R-AGILE-24-04-03-A",
            period_from=period_from,
            period_to=period_to,
        )

        expected_endpoint = (
            "products/AGILE-24-04-03/electricity-tariffs/E-1R-AGILE-24-04-03-A/"
            "standard-unit-rates/"
        )
        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "period_from": "2024-01-01T00:00:00+00:00",
                "period_to": "2024-01-31T00:00:00+00:00",
            },
        )
        assert len(result) == 1

    def test_get_electricity_standard_unit_rates_with_page_size(
        self, products_api: ProductsAPI, mock_charge_data: dict
    ) -> None:
        """Test getting electricity standard unit rates with page size."""
        products_api.client.get_paginated.return_value = [mock_charge_data]  # type: ignore[attr-defined]

        result = products_api.get_electricity_standard_unit_rates(
            "AGILE-24-04-03", "E-1R-AGILE-24-04-03-A", page_size=1000
        )

        expected_endpoint = (
            "products/AGILE-24-04-03/electricity-tariffs/E-1R-AGILE-24-04-03-A/"
            "standard-unit-rates/"
        )
        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {"page_size": 1000}
        )
        assert len(result) == 1


class TestElectricityDayUnitRates:
    """Tests for electricity day unit rates method."""

    def test_get_electricity_day_unit_rates(
        self, products_api: ProductsAPI, mock_charge_data: dict
    ) -> None:
        """Test getting electricity day unit rates."""
        products_api.client.get_paginated.return_value = [mock_charge_data]  # type: ignore[attr-defined]

        result = products_api.get_electricity_day_unit_rates(
            "VAR-24-04-03", "E-2R-VAR-24-04-03-A"
        )

        expected_endpoint = (
            "products/VAR-24-04-03/electricity-tariffs/E-2R-VAR-24-04-03-A/"
            "day-unit-rates/"
        )
        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {}
        )
        assert len(result) == 1
        assert isinstance(result[0], HistoricalCharge)


class TestElectricityNightUnitRates:
    """Tests for electricity night unit rates method."""

    def test_get_electricity_night_unit_rates(
        self, products_api: ProductsAPI, mock_charge_data: dict
    ) -> None:
        """Test getting electricity night unit rates."""
        products_api.client.get_paginated.return_value = [mock_charge_data]  # type: ignore[attr-defined]

        result = products_api.get_electricity_night_unit_rates(
            "VAR-24-04-03", "E-2R-VAR-24-04-03-A"
        )

        expected_endpoint = (
            "products/VAR-24-04-03/electricity-tariffs/E-2R-VAR-24-04-03-A/"
            "night-unit-rates/"
        )
        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {}
        )
        assert len(result) == 1
        assert isinstance(result[0], HistoricalCharge)


class TestElectricityStandingCharges:
    """Tests for electricity standing charges method."""

    def test_get_electricity_standing_charges(
        self, products_api: ProductsAPI, mock_charge_data: dict
    ) -> None:
        """Test getting electricity standing charges."""
        products_api.client.get_paginated.return_value = [mock_charge_data]  # type: ignore[attr-defined]

        result = products_api.get_electricity_standing_charges(
            "AGILE-24-04-03", "E-1R-AGILE-24-04-03-A"
        )

        expected_endpoint = (
            "products/AGILE-24-04-03/electricity-tariffs/E-1R-AGILE-24-04-03-A/"
            "standing-charges/"
        )
        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {}
        )
        assert len(result) == 1
        assert isinstance(result[0], HistoricalCharge)


class TestGasStandardUnitRates:
    """Tests for gas standard unit rates method."""

    def test_get_gas_standard_unit_rates(
        self, products_api: ProductsAPI, mock_charge_data: dict
    ) -> None:
        """Test getting gas standard unit rates."""
        products_api.client.get_paginated.return_value = [mock_charge_data]  # type: ignore[attr-defined]

        result = products_api.get_gas_standard_unit_rates(
            "VAR-24-04-03", "G-1R-VAR-24-04-03-A"
        )

        expected_endpoint = (
            "products/VAR-24-04-03/gas-tariffs/G-1R-VAR-24-04-03-A/standard-unit-rates/"
        )
        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {}
        )
        assert len(result) == 1
        assert isinstance(result[0], HistoricalCharge)


class TestGasStandingCharges:
    """Tests for gas standing charges method."""

    def test_get_gas_standing_charges(
        self, products_api: ProductsAPI, mock_charge_data: dict
    ) -> None:
        """Test getting gas standing charges."""
        products_api.client.get_paginated.return_value = [mock_charge_data]  # type: ignore[attr-defined]

        result = products_api.get_gas_standing_charges(
            "VAR-24-04-03", "G-1R-VAR-24-04-03-A"
        )

        expected_endpoint = (
            "products/VAR-24-04-03/gas-tariffs/G-1R-VAR-24-04-03-A/standing-charges/"
        )
        products_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {}
        )
        assert len(result) == 1
        assert isinstance(result[0], HistoricalCharge)



"""Tests for exceptions."""

from pytentacle.exceptions import (
    APIError,
    AuthenticationError,
    NotFoundError,
    PytentacleError,
    RateLimitError,
    ServerError,
    ValidationError,
)


def test_base_exception() -> None:
    """Test base PytentacleError."""
    error = PytentacleError("Test error")
    assert str(error) == "Test error"


def test_api_error() -> None:
    """Test APIError with status code."""
    error = APIError("API error", status_code=400)
    assert str(error) == "API error"
    assert error.status_code == 400


def test_authentication_error() -> None:
    """Test AuthenticationError."""
    error = AuthenticationError("Auth failed", status_code=401)
    assert str(error) == "Auth failed"
    assert error.status_code == 401
    assert isinstance(error, APIError)


def test_validation_error() -> None:
    """Test ValidationError."""
    error = ValidationError("Validation failed", status_code=400)
    assert isinstance(error, APIError)


def test_not_found_error() -> None:
    """Test NotFoundError."""
    error = NotFoundError("Not found", status_code=404)
    assert isinstance(error, APIError)


def test_rate_limit_error() -> None:
    """Test RateLimitError."""
    error = RateLimitError("Rate limited", status_code=429)
    assert isinstance(error, APIError)


def test_server_error() -> None:
    """Test ServerError."""
    error = ServerError("Server error", status_code=500)
    assert isinstance(error, APIError)



"""Pytest configuration and fixtures."""

import pytest

from pytentacle import AsyncOctopusClient, OctopusClient


@pytest.fixture
def api_key() -> str:
    """Return a test API key."""
    return "sk_test_1234567890"


@pytest.fixture
def client(api_key: str) -> OctopusClient:
    """Return a test client."""
    return OctopusClient(api_key=api_key)


@pytest.fixture
async def async_client(api_key: str) -> AsyncOctopusClient:
    """Return an async test client."""
    return AsyncOctopusClient(api_key=api_key)



"""Tests for Electricity API endpoints."""

from datetime import datetime, timezone
from unittest.mock import MagicMock

import pytest

from pytentacle.api.electricity import ElectricityAPI
from pytentacle.models.consumption import Consumption
from pytentacle.models.meter_points import ElectricityMeterPoint


@pytest.fixture
def electricity_api() -> ElectricityAPI:
    """Return an electricity API instance with mocked client."""
    client = MagicMock()
    return ElectricityAPI(client)


@pytest.fixture
def mock_meter_point_data() -> dict:
    """Return mock electricity meter point data."""
    return {
        "mpan": "1234567890123",
        "profile_class": 1,
        "consumption_standard": 2900,
        "gsp": "_A",
    }


@pytest.fixture
def mock_consumption_data() -> dict:
    """Return mock consumption data."""
    return {
        "consumption": 0.5,
        "interval_start": "2024-01-01T00:00:00Z",
        "interval_end": "2024-01-01T00:30:00Z",
    }


class TestGetMeterPoint:
    """Tests for get_meter_point method."""

    def test_get_meter_point(
        self, electricity_api: ElectricityAPI, mock_meter_point_data: dict
    ) -> None:
        """Test getting an electricity meter point."""
        electricity_api.client.get.return_value = mock_meter_point_data  # type: ignore[attr-defined]

        result = electricity_api.get_meter_point("1234567890123")

        electricity_api.client.get.assert_called_once_with(
            "electricity-meter-points/1234567890123/"
        )  # type: ignore[attr-defined]
        assert isinstance(result, ElectricityMeterPoint)
        assert result.mpan == "1234567890123"


class TestGetConsumption:
    """Tests for get_consumption method."""

    def test_get_consumption_minimal(
        self, electricity_api: ElectricityAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with minimal parameters."""
        electricity_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]

        result = electricity_api.get_consumption("1234567890123", "12A3456789")

        expected_endpoint = (
            "electricity-meter-points/1234567890123/meters/12A3456789/consumption/"
        )
        electricity_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {"order_by": "-period"}
        )
        assert len(result) == 1
        assert isinstance(result[0], Consumption)
        assert result[0].consumption == 0.5

    def test_get_consumption_with_dates(
        self, electricity_api: ElectricityAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with date filters."""
        electricity_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]
        period_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
        period_to = datetime(2024, 1, 31, tzinfo=timezone.utc)

        result = electricity_api.get_consumption(
            "1234567890123",
            "12A3456789",
            period_from=period_from,
            period_to=period_to,
        )

        expected_endpoint = (
            "electricity-meter-points/1234567890123/meters/12A3456789/consumption/"
        )
        electricity_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "-period",
                "period_from": "2024-01-01T00:00:00+00:00",
                "period_to": "2024-01-31T00:00:00+00:00",
            },
        )
        assert len(result) == 1

    def test_get_consumption_with_page_size(
        self, electricity_api: ElectricityAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with page size."""
        electricity_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]

        result = electricity_api.get_consumption(
            "1234567890123",
            "12A3456789",
            page_size=5000,
        )

        expected_endpoint = (
            "electricity-meter-points/1234567890123/meters/12A3456789/consumption/"
        )
        electricity_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "-period",
                "page_size": 5000,
            },
        )
        assert len(result) == 1

    def test_get_consumption_with_order_by_asc(
        self, electricity_api: ElectricityAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with ascending order."""
        electricity_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]

        result = electricity_api.get_consumption(
            "1234567890123",
            "12A3456789",
            order_by="period",
        )

        expected_endpoint = (
            "electricity-meter-points/1234567890123/meters/12A3456789/consumption/"
        )
        electricity_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "period",
            },
        )
        assert len(result) == 1

    def test_get_consumption_with_group_by(
        self, electricity_api: ElectricityAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption grouped by day."""
        electricity_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]

        result = electricity_api.get_consumption(
            "1234567890123",
            "12A3456789",
            group_by="day",
        )

        expected_endpoint = (
            "electricity-meter-points/1234567890123/meters/12A3456789/consumption/"
        )
        electricity_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "-period",
                "group_by": "day",
            },
        )
        assert len(result) == 1

    def test_get_consumption_with_all_params(
        self, electricity_api: ElectricityAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with all parameters."""
        electricity_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]
        period_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
        period_to = datetime(2024, 1, 31, tzinfo=timezone.utc)

        result = electricity_api.get_consumption(
            "1234567890123",
            "12A3456789",
            period_from=period_from,
            period_to=period_to,
            page_size=10000,
            order_by="period",
            group_by="week",
        )

        expected_endpoint = (
            "electricity-meter-points/1234567890123/meters/12A3456789/consumption/"
        )
        electricity_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "period",
                "period_from": "2024-01-01T00:00:00+00:00",
                "period_to": "2024-01-31T00:00:00+00:00",
                "page_size": 10000,
                "group_by": "week",
            },
        )
        assert len(result) == 1



"""Tests for AsyncBaseClient."""

from unittest.mock import AsyncMock, MagicMock, patch

import httpx
import pytest

from pytentacle.core.async_base import AsyncBaseClient
from pytentacle.exceptions import (
    AuthenticationError,
    NotFoundError,
    RateLimitError,
    ServerError,
    ValidationError,
)


@pytest.fixture
def async_base_client() -> AsyncBaseClient:
    """Return an async base client instance."""
    return AsyncBaseClient(api_key="test_key")


@pytest.fixture
def mock_response() -> MagicMock:
    """Return a mock HTTP response."""
    response = MagicMock(spec=httpx.Response)
    response.status_code = 200
    response.json.return_value = {"test": "data"}
    return response


class TestAsyncBaseClientInit:
    """Tests for AsyncBaseClient initialization."""

    def test_init_with_api_key(self) -> None:
        """Test initialization with API key."""
        client = AsyncBaseClient(api_key="test_key")
        assert client.base_url == "https://api.octopus.energy/v1"
        assert client.timeout == 30.0
        assert client.auth is not None

    def test_init_without_api_key(self) -> None:
        """Test initialization without API key."""
        client = AsyncBaseClient()
        assert client.auth is None

    def test_init_with_custom_base_url(self) -> None:
        """Test initialization with custom base URL."""
        client = AsyncBaseClient(base_url="https://custom.url")
        assert client.base_url == "https://custom.url"

    def test_init_with_custom_timeout(self) -> None:
        """Test initialization with custom timeout."""
        client = AsyncBaseClient(timeout=60.0)
        assert client.timeout == 60.0


class TestAsyncBaseClientProperty:
    """Tests for AsyncBaseClient client property."""

    def test_client_property_creates_client(
        self, async_base_client: AsyncBaseClient
    ) -> None:
        """Test that client property creates httpx.AsyncClient."""
        client = async_base_client.client
        assert isinstance(client, httpx.AsyncClient)

    def test_client_property_reuses_client(
        self, async_base_client: AsyncBaseClient
    ) -> None:
        """Test that client property reuses the same client instance."""
        client1 = async_base_client.client
        client2 = async_base_client.client
        assert client1 is client2


class TestAsyncGetHeaders:
    """Tests for _get_headers method."""

    def test_get_headers_with_auth(self, async_base_client: AsyncBaseClient) -> None:
        """Test getting headers with authentication."""
        headers = async_base_client._get_headers()
        assert "Content-Type" in headers
        assert "Authorization" in headers

    def test_get_headers_without_auth(self) -> None:
        """Test getting headers without authentication."""
        client = AsyncBaseClient()
        headers = client._get_headers()
        assert "Content-Type" in headers
        assert "Authorization" not in headers


class TestAsyncHandleResponse:
    """Tests for _handle_response method."""

    def test_handle_response_success(
        self, async_base_client: AsyncBaseClient, mock_response: MagicMock
    ) -> None:
        """Test handling successful response."""
        result = async_base_client._handle_response(mock_response)
        assert result == {"test": "data"}

    def test_handle_response_401(self, async_base_client: AsyncBaseClient) -> None:
        """Test handling 401 response."""
        response = MagicMock(spec=httpx.Response)
        response.status_code = 401

        with pytest.raises(AuthenticationError) as exc_info:
            async_base_client._handle_response(response)
        assert exc_info.value.status_code == 401

    def test_handle_response_404(self, async_base_client: AsyncBaseClient) -> None:
        """Test handling 404 response."""
        response = MagicMock(spec=httpx.Response)
        response.status_code = 404

        with pytest.raises(NotFoundError) as exc_info:
            async_base_client._handle_response(response)
        assert exc_info.value.status_code == 404

    def test_handle_response_400(self, async_base_client: AsyncBaseClient) -> None:
        """Test handling 400 response."""
        response = MagicMock(spec=httpx.Response)
        response.status_code = 400
        response.json.return_value = {"detail": "Invalid request"}

        with pytest.raises(ValidationError) as exc_info:
            async_base_client._handle_response(response)
        assert exc_info.value.status_code == 400

    def test_handle_response_400_no_detail(
        self, async_base_client: AsyncBaseClient
    ) -> None:
        """Test handling 400 response without detail."""
        response = MagicMock(spec=httpx.Response)
        response.status_code = 400
        response.json.side_effect = Exception("No JSON")

        with pytest.raises(ValidationError) as exc_info:
            async_base_client._handle_response(response)
        assert "Validation error" in str(exc_info.value)

    def test_handle_response_429(self, async_base_client: AsyncBaseClient) -> None:
        """Test handling 429 response."""
        response = MagicMock(spec=httpx.Response)
        response.status_code = 429

        with pytest.raises(RateLimitError) as exc_info:
            async_base_client._handle_response(response)
        assert exc_info.value.status_code == 429

    def test_handle_response_500(self, async_base_client: AsyncBaseClient) -> None:
        """Test handling 500 response."""
        response = MagicMock(spec=httpx.Response)
        response.status_code = 500

        with pytest.raises(ServerError) as exc_info:
            async_base_client._handle_response(response)
        assert exc_info.value.status_code == 500

    def test_handle_response_503(self, async_base_client: AsyncBaseClient) -> None:
        """Test handling 503 response."""
        response = MagicMock(spec=httpx.Response)
        response.status_code = 503

        with pytest.raises(ServerError) as exc_info:
            async_base_client._handle_response(response)
        assert exc_info.value.status_code == 503


class TestAsyncGet:
    """Tests for get method."""

    @pytest.mark.asyncio
    async def test_get_success(self, async_base_client: AsyncBaseClient) -> None:
        """Test successful GET request."""
        mock_response = MagicMock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": "test"}

        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=mock_response)

        async_base_client._client = mock_client

        result = await async_base_client.get("test-endpoint")

        assert result == {"data": "test"}
        mock_client.get.assert_called_once()

    @pytest.mark.asyncio
    async def test_get_with_params(self, async_base_client: AsyncBaseClient) -> None:
        """Test GET request with parameters."""
        mock_response = MagicMock(spec=httpx.Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {"data": "test"}

        mock_client = AsyncMock()
        mock_client.get = AsyncMock(return_value=mock_response)

        async_base_client._client = mock_client

        result = await async_base_client.get("test-endpoint", {"param": "value"})

        assert result == {"data": "test"}
        call_args = mock_client.get.call_args
        assert call_args.kwargs["params"] == {"param": "value"}


class TestAsyncGetPaginated:
    """Tests for get_paginated method."""

    @pytest.mark.asyncio
    async def test_get_paginated_single_page(
        self, async_base_client: AsyncBaseClient
    ) -> None:
        """Test paginated GET with single page."""
        mock_response = {
            "results": [{"id": 1}, {"id": 2}],
            "next": None,
        }

        with patch.object(async_base_client, "get", new_callable=AsyncMock) as mock_get:
            mock_get.return_value = mock_response

            result = await async_base_client.get_paginated("test-endpoint")

            assert len(result) == 2
            assert result[0]["id"] == 1
            assert result[1]["id"] == 2

    @pytest.mark.asyncio
    async def test_get_paginated_multiple_pages(
        self, async_base_client: AsyncBaseClient
    ) -> None:
        """Test paginated GET with multiple pages."""
        page1 = {
            "results": [{"id": 1}],
            "next": "https://api.test.com/endpoint?page=2",
        }
        page2 = {
            "results": [{"id": 2}],
            "next": None,
        }

        with patch.object(async_base_client, "get", new_callable=AsyncMock) as mock_get:
            mock_get.side_effect = [page1, page2]

            result = await async_base_client.get_paginated("test-endpoint")

            assert len(result) == 2
            assert result[0]["id"] == 1
            assert result[1]["id"] == 2

    @pytest.mark.asyncio
    async def test_get_paginated_non_paginated_response(
        self, async_base_client: AsyncBaseClient
    ) -> None:
        """Test paginated GET with non-paginated response."""
        mock_response = {"id": 1, "name": "test"}

        with patch.object(async_base_client, "get", new_callable=AsyncMock) as mock_get:
            mock_get.return_value = mock_response

            result = await async_base_client.get_paginated("test-endpoint")

            assert len(result) == 1
            assert result[0] == mock_response


class TestAsyncClose:
    """Tests for close method."""

    @pytest.mark.asyncio
    async def test_close(self, async_base_client: AsyncBaseClient) -> None:
        """Test closing the client."""
        # Create the client first
        _ = async_base_client.client

        await async_base_client.close()
        assert async_base_client._client is None


class TestAsyncContextManager:
    """Tests for async context manager."""

    @pytest.mark.asyncio
    async def test_context_manager(self) -> None:
        """Test using client as async context manager."""
        async with AsyncBaseClient(api_key="test_key") as client:
            assert client is not None
            assert isinstance(client, AsyncBaseClient)



"""Industry data API endpoints."""

from typing import TYPE_CHECKING

from pytentacle.models.common import GridSupplyPoint

if TYPE_CHECKING:
    from pytentacle.core.base import BaseClient


class IndustryAPI:
    """Industry data API endpoints.

    Provides access to Grid Supply Point (GSP) data.
    GSPs are the 14 regional distribution zones in the UK.
    """

    def __init__(self, client: "BaseClient") -> None:
        """Initialize the industry API.

        Args:
            client: The HTTP client to use
        """
        self.client = client

    def get_grid_supply_points(
        self,
        postcode: str | None = None,
    ) -> list[GridSupplyPoint]:
        """Get list of Grid Supply Points, optionally filtered by postcode.

        Args:
            postcode: Optional UK postcode to filter by

        Returns:
            List of GridSupplyPoint objects
        """
        params = {}
        if postcode:
            params["postcode"] = postcode

        results = self.client.get_paginated("industry/grid-supply-points/", params)
        return [GridSupplyPoint.model_validate(item) for item in results]



"""Common models used across the API."""

from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict, Field

T = TypeVar("T")


class Link(BaseModel):
    """HATEOAS link."""

    href: str = Field(description="URL of the linked resource")
    method: str = Field(description="HTTP method to use")
    rel: str = Field(description="Relationship type")


class PaginatedResponse(BaseModel, Generic[T]):
    """Paginated response wrapper."""

    count: int = Field(description="Total number of results")
    next: str | None = Field(None, description="URL to next page")
    previous: str | None = Field(None, description="URL to previous page")
    results: list[T] = Field(description="List of results for this page")


class GridSupplyPoint(BaseModel):
    """Grid Supply Point (GSP) identifier.

    GSPs are the 14 regional distribution zones in the UK.
    """

    group_id: str = Field(
        description="GSP group ID (e.g., '_A', '_B', '_H')",
        alias="groupId",
    )

    model_config = ConfigDict(populate_by_name=True)



"""Tests for product models."""

from datetime import datetime

from pytentacle.models.products import Product


def test_product_model() -> None:
    """Test Product model validation."""
    data = {
        "code": "AGILE-24-04-03",
        "direction": "IMPORT",
        "full_name": "Agile Octopus April 2024",
        "display_name": "Agile Octopus",
        "description": "Variable tariff tracking wholesale prices",
        "is_variable": True,
        "is_green": True,
        "is_tracker": True,
        "is_prepay": False,
        "is_business": False,
        "is_restricted": False,
        "term": 12,
        "available_from": "2024-04-03T00:00:00Z",
        "available_to": None,
        "links": [],
        "brand": "OCTOPUS_ENERGY",
    }

    product = Product.model_validate(data)

    assert product.code == "AGILE-24-04-03"
    assert product.is_variable is True
    assert product.is_green is True
    assert product.brand == "OCTOPUS_ENERGY"
    assert isinstance(product.available_from, datetime)
    assert product.available_to is None



"""Electricity meter points API endpoints."""

import os
from datetime import datetime
from typing import TYPE_CHECKING, Literal

from pytentacle.models.consumption import Consumption
from pytentacle.models.meter_points import ElectricityMeterPoint

if TYPE_CHECKING:
    from pytentacle.core.base import BaseClient


class ElectricityAPI:
    """Electricity meter points API endpoints.

    Provides access to electricity meter point details and consumption data.
    Note: Half-hourly consumption data is only available for smart meters.
    """

    def __init__(self, client: "BaseClient") -> None:
        """Initialize the electricity API.

        Args:
            client: The HTTP client to use
        """
        self.client = client

    def get_meter_point(self, mpan: str | None = None) -> ElectricityMeterPoint:
        """Get electricity meter point details.

        Args:
            mpan: Meter Point Administration Number (13 digits). If not provided,
                will attempt to read from OCTOPUS_ELECTRICITY_MPAN environment variable.

        Returns:
            ElectricityMeterPoint object

        Raises:
            ValueError: If mpan is not provided and not found in environment
        """
        if mpan is None:
            mpan = os.getenv("OCTOPUS_ELECTRICITY_MPAN")
            if mpan is None:
                msg = (
                    "mpan is required. Provide it as an argument or set "
                    "OCTOPUS_ELECTRICITY_MPAN environment variable."
                )
                raise ValueError(msg)

        data = self.client.get(f"electricity-meter-points/{mpan}/")
        return ElectricityMeterPoint.model_validate(data)

    def get_consumption(  # noqa: PLR0913
        self,
        mpan: str | None = None,
        serial_number: str | None = None,
        *,
        period_from: datetime | None = None,
        period_to: datetime | None = None,
        page_size: int | None = None,
        order_by: Literal["period", "-period"] = "-period",
        group_by: Literal["hour", "day", "week", "month", "quarter"] | None = None,
    ) -> list[Consumption]:
        """Get electricity consumption data for a meter.

        Half-hourly consumption data is only available for smart meters.
        Non-smart meter requests will return empty results.

        Args:
            mpan: Meter Point Administration Number (13 digits). If not provided,
                will attempt to read from OCTOPUS_ELECTRICITY_MPAN
                environment variable.
            serial_number: Electricity meter serial number. If not provided,
                will attempt to read from OCTOPUS_ELECTRICITY_SERIAL_NUMBER
                environment variable.
            period_from: Filter consumption from this datetime (inclusive)
            period_to: Filter consumption to this datetime (exclusive)
            page_size: Number of results per page (max 25000)
            order_by: Sort order - 'period' (ascending) or '-period' (descending)
            group_by: Group results by time period

        Returns:
            List of Consumption objects

        Raises:
            ValueError: If mpan or serial_number is not provided and not found
                in environment
        """
        if mpan is None:
            mpan = os.getenv("OCTOPUS_ELECTRICITY_MPAN")
            if mpan is None:
                msg = (
                    "mpan is required. Provide it as an argument or set "
                    "OCTOPUS_ELECTRICITY_MPAN environment variable."
                )
                raise ValueError(msg)

        if serial_number is None:
            serial_number = os.getenv("OCTOPUS_ELECTRICITY_SERIAL_NUMBER")
            if serial_number is None:
                msg = (
                    "serial_number is required. Provide it as an argument or "
                    "set OCTOPUS_ELECTRICITY_SERIAL_NUMBER environment variable."
                )
                raise ValueError(msg)
        params: dict[str, str | int] = {"order_by": order_by}
        if period_from:
            params["period_from"] = period_from.isoformat()
        if period_to:
            params["period_to"] = period_to.isoformat()
        if page_size:
            params["page_size"] = page_size
        if group_by:
            params["group_by"] = group_by

        endpoint = (
            f"electricity-meter-points/{mpan}/meters/{serial_number}/consumption/"
        )
        results = self.client.get_paginated(endpoint, params)
        return [Consumption.model_validate(item) for item in results]



"""Pydantic models for API data structures."""

from pytentacle.models.common import GridSupplyPoint, Link, PaginatedResponse
from pytentacle.models.consumption import Consumption
from pytentacle.models.meter_points import ElectricityMeterPoint
from pytentacle.models.products import Product
from pytentacle.models.tariffs import (
    Eco7ElectricityTariff,
    GasTariff,
    HistoricalCharge,
    StandardElectricityTariff,
)

__all__ = [
    "Consumption",
    "Eco7ElectricityTariff",
    "ElectricityMeterPoint",
    "GasTariff",
    "GridSupplyPoint",
    "HistoricalCharge",
    "Link",
    "PaginatedResponse",
    "Product",
    "StandardElectricityTariff",
]



"""Tests for authentication."""

from pytentacle.auth import APIKeyAuth


def test_api_key_auth() -> None:
    """Test API key authentication."""
    auth = APIKeyAuth("test_key_12345")
    headers = auth.get_headers()

    assert "Authorization" in headers
    assert headers["Authorization"] == "Basic dGVzdF9rZXlfMTIzNDU6"


def test_api_key_auth_repr() -> None:
    """Test API key auth string representation."""
    auth = APIKeyAuth("test_key_12345")
    repr_str = repr(auth)

    assert "***2345" in repr_str
    assert "test_key" not in repr_str  # Full key should be hidden



"""Tests for consumption models."""

from datetime import datetime

from pytentacle.models.consumption import Consumption


def test_consumption_model() -> None:
    """Test Consumption model validation."""
    data = {
        "consumption": 0.123,
        "interval_start": "2024-01-01T00:00:00Z",
        "interval_end": "2024-01-01T00:30:00Z",
    }

    consumption = Consumption.model_validate(data)

    assert consumption.consumption == 0.123
    assert isinstance(consumption.interval_start, datetime)
    assert isinstance(consumption.interval_end, datetime)



"""Base HTTP client for synchronous requests."""

from typing import Any

import httpx

from pytentacle.auth import APIKeyAuth
from pytentacle.exceptions import (
    AuthenticationError,
    NotFoundError,
    RateLimitError,
    ServerError,
    ValidationError,
)
from pytentacle.types import Headers, JSONDict


class BaseClient:
    """Base HTTP client for the Octopus Energy API."""

    BASE_URL = "https://api.octopus.energy/v1"

    def __init__(
        self,
        api_key: str | None = None,
        base_url: str | None = None,
        timeout: float = 30.0,
    ) -> None:
        """Initialize the base client.

        Args:
            api_key: Optional API key for authentication
            base_url: Optional base URL (defaults to production API)
            timeout: Request timeout in seconds
        """
        self.base_url = base_url or self.BASE_URL
        self.timeout = timeout
        self.auth = APIKeyAuth(api_key) if api_key else None
        self._client: httpx.Client | None = None

    @property
    def client(self) -> httpx.Client:
        """Get or create the HTTP client.

        Returns:
            httpx.Client instance
        """
        if self._client is None:
            self._client = httpx.Client(timeout=self.timeout)
        return self._client

    def _get_headers(self) -> Headers:
        """Get request headers including authentication.

        Returns:
            Dictionary of headers
        """
        headers: Headers = {"Content-Type": "application/json"}
        if self.auth:
            headers.update(self.auth.get_headers())
        return headers

    def _handle_response(self, response: httpx.Response) -> JSONDict:
        """Handle API response and raise appropriate exceptions.

        Args:
            response: The HTTP response

        Returns:
            Parsed JSON response

        Raises:
            AuthenticationError: For 401 responses
            NotFoundError: For 404 responses
            ValidationError: For 400 responses
            RateLimitError: For 429 responses
            ServerError: For 5xx responses
        """
        if response.status_code == 401:
            msg = "Authentication failed. Check your API key."
            raise AuthenticationError(
                msg,
                status_code=401,
            )
        if response.status_code == 404:
            msg = "Resource not found."
            raise NotFoundError(
                msg,
                status_code=404,
            )
        if response.status_code == 400:
            try:
                error_data = response.json()
                message = error_data.get("detail", "Validation error")
            except Exception:  # noqa: BLE001
                message = "Validation error"
            raise ValidationError(message, status_code=400)
        if response.status_code == 429:
            msg = "Rate limit exceeded. Please try again later."
            raise RateLimitError(
                msg,
                status_code=429,
            )
        if response.status_code >= 500:
            msg = f"Server error: {response.status_code}"
            raise ServerError(
                msg,
                status_code=response.status_code,
            )

        response.raise_for_status()
        return response.json()

    def get(self, endpoint: str, params: dict[str, Any] | None = None) -> JSONDict:
        """Make a GET request.

        Args:
            endpoint: API endpoint (relative to base URL)
            params: Optional query parameters

        Returns:
            JSON response as dictionary
        """
        url = f"{self.base_url}/{endpoint.lstrip('/')}"
        response = self.client.get(url, params=params, headers=self._get_headers())
        return self._handle_response(response)

    def get_paginated(
        self,
        endpoint: str,
        params: dict[str, Any] | None = None,
    ) -> list[JSONDict]:
        """Make a GET request and handle pagination automatically.

        Args:
            endpoint: API endpoint (relative to base URL)
            params: Optional query parameters

        Returns:
            List of all results from all pages
        """
        all_results: list[JSONDict] = []
        current_params = params.copy() if params else {}

        while True:
            data = self.get(endpoint, current_params)

            # Check if response has pagination structure
            if isinstance(data, dict) and "results" in data:
                all_results.extend(data["results"])

                # Check for next page
                if data.get("next"):
                    # Extract page number from next URL
                    next_url = data["next"]
                    if "page=" in next_url:
                        page_num = next_url.split("page=")[1].split("&")[0]
                        current_params["page"] = page_num
                    else:
                        break
                else:
                    break
            else:
                # Not a paginated response, return as-is
                return [data]

        return all_results

    def close(self) -> None:
        """Close the HTTP client."""
        if self._client is not None:
            self._client.close()
            self._client = None

    def __enter__(self) -> "BaseClient":
        """Context manager entry."""
        return self

    def __exit__(self, *args: object) -> None:
        """Context manager exit."""
        self.close()



"""Tests for Industry API endpoints."""

from unittest.mock import MagicMock

import pytest

from pytentacle.api.industry import IndustryAPI
from pytentacle.models.common import GridSupplyPoint


@pytest.fixture
def industry_api() -> IndustryAPI:
    """Return an industry API instance with mocked client."""
    client = MagicMock()
    return IndustryAPI(client)


@pytest.fixture
def mock_gsp_data() -> dict:
    """Return mock Grid Supply Point data."""
    return {
        "groupId": "_A",
    }


class TestGetGridSupplyPoints:
    """Tests for get_grid_supply_points method."""

    def test_get_grid_supply_points_no_filter(
        self, industry_api: IndustryAPI, mock_gsp_data: dict
    ) -> None:
        """Test getting all grid supply points without filter."""
        industry_api.client.get_paginated.return_value = [mock_gsp_data]  # type: ignore[attr-defined]

        result = industry_api.get_grid_supply_points()

        industry_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "industry/grid-supply-points/", {}
        )
        assert len(result) == 1
        assert isinstance(result[0], GridSupplyPoint)
        assert result[0].group_id == "_A"

    def test_get_grid_supply_points_with_postcode(
        self, industry_api: IndustryAPI, mock_gsp_data: dict
    ) -> None:
        """Test getting grid supply points filtered by postcode."""
        industry_api.client.get_paginated.return_value = [mock_gsp_data]  # type: ignore[attr-defined]

        result = industry_api.get_grid_supply_points(postcode="SW1A 1AA")

        industry_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            "industry/grid-supply-points/", {"postcode": "SW1A 1AA"}
        )
        assert len(result) == 1
        assert isinstance(result[0], GridSupplyPoint)



"""Product models."""

from datetime import datetime

from pydantic import BaseModel, Field

from pytentacle.models.tariffs import (
    Eco7ElectricityTariff,
    GasTariff,
    StandardElectricityTariff,
)


class Product(BaseModel):
    """Energy product with associated tariffs.

    Products have multiple tariff variants based on:
    - GSP (Grid Supply Point) - 14 regional zones in UK
    - Payment method (direct debit, prepay, etc.)
    - Register type (single vs dual for electricity)
    """

    code: str = Field(description="Product code (e.g., 'AGILE-24-04-03')")
    direction: str | None = Field(
        None,
        description="Direction (IMPORT for consumption, EXPORT for generation)",
    )
    full_name: str = Field(description="Full product name")
    display_name: str = Field(description="Display name")
    description: str = Field(description="Product description")
    is_variable: bool = Field(description="Whether the product has variable rates")
    is_green: bool = Field(description="Whether the product is 100% renewable")
    is_tracker: bool = Field(description="Whether the product tracks wholesale prices")
    is_prepay: bool = Field(description="Whether the product is prepayment")
    is_business: bool = Field(description="Whether the product is for business")
    is_restricted: bool = Field(
        description="Whether the product has restricted availability",
    )
    term: int | None = Field(None, description="Contract term in months")
    available_from: datetime = Field(description="Date product became available")
    available_to: datetime | None = Field(
        None,
        description="Date product is available until",
    )
    links: list[dict[str, str]] = Field(
        default_factory=list,
        description="HATEOAS links",
    )
    brand: str = Field(description="Brand name")

    # Tariffs - these are populated when fetching a specific product
    # Structure: dict with GSP keys, each containing payment_method: tariff_data
    single_register_electricity_tariffs: dict[
        str,
        dict[str, StandardElectricityTariff],
    ] = Field(
        default_factory=dict,
        description="Single register electricity tariffs by GSP and payment method",
    )
    dual_register_electricity_tariffs: dict[str, dict[str, Eco7ElectricityTariff]] = (
        Field(
            default_factory=dict,
            description=(
                "Dual register (Economy 7) electricity tariffs by GSP "
                "and payment method"
            ),
        )
    )
    single_register_gas_tariffs: dict[str, dict[str, GasTariff]] = Field(
        default_factory=dict,
        description="Gas tariffs by GSP and payment method",
    )
    sample_quotes: dict[str, dict[str, dict[str, dict[str, int]]]] | None = Field(
        None,
        description=(
            "Sample cost calculations by GSP, payment method, and consumption type"
        ),
    )
    sample_consumption: dict[str, dict[str, int]] | None = Field(
        None,
        description="Standard consumption values used for quotes",
    )



"""Authentication for the Octopus Energy API."""

import base64

from pytentacle.types import Headers


class APIKeyAuth:
    """API Key authentication for Octopus Energy API.

    The API uses Basic Authentication with the API key as username and empty password.
    """

    def __init__(self, api_key: str) -> None:
        """Initialize API key authentication.

        Args:
            api_key: The API key from Octopus Energy
        """
        self.api_key = api_key

    def get_headers(self) -> Headers:
        """Get authentication headers.

        Returns:
            Dictionary containing the Authorization header
        """
        credentials = f"{self.api_key}:"
        encoded = base64.b64encode(credentials.encode()).decode()
        return {"Authorization": f"Basic {encoded}"}

    def __repr__(self) -> str:
        """Return string representation."""
        return f"APIKeyAuth(api_key='***{self.api_key[-4:]}')"



"""Exceptions for the Pytentacle library."""


class PytentacleError(Exception):
    """Base exception for all Pytentacle errors."""


class APIError(PytentacleError):
    """Raised when the API returns an error response."""

    def __init__(self, message: str, status_code: int | None = None) -> None:
        """Initialize the APIError.

        Args:
            message: Error message from the API
            status_code: HTTP status code
        """
        super().__init__(message)
        self.status_code = status_code


class AuthenticationError(APIError):
    """Raised when authentication fails."""


class ValidationError(APIError):
    """Raised when request validation fails."""


class NotFoundError(APIError):
    """Raised when a resource is not found."""


class RateLimitError(APIError):
    """Raised when rate limit is exceeded."""


class ServerError(APIError):
    """Raised when the server returns a 5xx error."""



"""Tests for the main client."""

from unittest.mock import patch

from pytentacle import OctopusClient


def test_client_initialization() -> None:
    """Test OctopusClient initialization."""
    client = OctopusClient(api_key="test_key")

    assert client.products is not None
    assert client.electricity is not None
    assert client.gas is not None
    assert client.industry is not None


def test_client_context_manager() -> None:
    """Test OctopusClient as context manager."""
    with OctopusClient(api_key="test_key") as client:
        assert client.products is not None

    assert client._base_client._client is None  # Should be closed


@patch.dict("os.environ", {}, clear=True)
def test_client_repr() -> None:
    """Test OctopusClient string representation."""
    client = OctopusClient(api_key="test_key")
    assert "authenticated" in repr(client)

    client_no_auth = OctopusClient()
    assert "unauthenticated" in repr(client_no_auth)



"""Tests for Gas API endpoints."""

from datetime import datetime, timezone
from unittest.mock import MagicMock

import pytest

from pytentacle.api.gas import GasAPI
from pytentacle.models.consumption import Consumption


@pytest.fixture
def gas_api() -> GasAPI:
    """Return a gas API instance with mocked client."""
    client = MagicMock()
    return GasAPI(client)


@pytest.fixture
def mock_consumption_data() -> dict:
    """Return mock consumption data."""
    return {
        "consumption": 10.5,
        "interval_start": "2024-01-01T00:00:00Z",
        "interval_end": "2024-01-01T00:30:00Z",
    }


class TestGetConsumption:
    """Tests for get_consumption method."""

    def test_get_consumption_minimal(
        self, gas_api: GasAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with minimal parameters."""
        gas_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]

        result = gas_api.get_consumption("1234567890", "G12A3456789")

        expected_endpoint = (
            "gas-meter-points/1234567890/meters/G12A3456789/consumption/"
        )
        gas_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint, {"order_by": "-period"}
        )
        assert len(result) == 1
        assert isinstance(result[0], Consumption)
        assert result[0].consumption == 10.5

    def test_get_consumption_with_dates(
        self, gas_api: GasAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with date filters."""
        gas_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]
        period_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
        period_to = datetime(2024, 1, 31, tzinfo=timezone.utc)

        result = gas_api.get_consumption(
            "1234567890",
            "G12A3456789",
            period_from=period_from,
            period_to=period_to,
        )

        expected_endpoint = (
            "gas-meter-points/1234567890/meters/G12A3456789/consumption/"
        )
        gas_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "-period",
                "period_from": "2024-01-01T00:00:00+00:00",
                "period_to": "2024-01-31T00:00:00+00:00",
            },
        )
        assert len(result) == 1

    def test_get_consumption_with_page_size(
        self, gas_api: GasAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with page size."""
        gas_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]

        result = gas_api.get_consumption(
            "1234567890",
            "G12A3456789",
            page_size=5000,
        )

        expected_endpoint = (
            "gas-meter-points/1234567890/meters/G12A3456789/consumption/"
        )
        gas_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "-period",
                "page_size": 5000,
            },
        )
        assert len(result) == 1

    def test_get_consumption_with_order_by_asc(
        self, gas_api: GasAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with ascending order."""
        gas_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]

        result = gas_api.get_consumption(
            "1234567890",
            "G12A3456789",
            order_by="period",
        )

        expected_endpoint = (
            "gas-meter-points/1234567890/meters/G12A3456789/consumption/"
        )
        gas_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "period",
            },
        )
        assert len(result) == 1

    def test_get_consumption_with_group_by(
        self, gas_api: GasAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption grouped by day."""
        gas_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]

        result = gas_api.get_consumption(
            "1234567890",
            "G12A3456789",
            group_by="day",
        )

        expected_endpoint = (
            "gas-meter-points/1234567890/meters/G12A3456789/consumption/"
        )
        gas_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "-period",
                "group_by": "day",
            },
        )
        assert len(result) == 1

    def test_get_consumption_with_all_params(
        self, gas_api: GasAPI, mock_consumption_data: dict
    ) -> None:
        """Test getting consumption with all parameters."""
        gas_api.client.get_paginated.return_value = [mock_consumption_data]  # type: ignore[attr-defined]
        period_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
        period_to = datetime(2024, 1, 31, tzinfo=timezone.utc)

        result = gas_api.get_consumption(
            "1234567890",
            "G12A3456789",
            period_from=period_from,
            period_to=period_to,
            page_size=10000,
            order_by="period",
            group_by="week",
        )

        expected_endpoint = (
            "gas-meter-points/1234567890/meters/G12A3456789/consumption/"
        )
        gas_api.client.get_paginated.assert_called_once_with(  # type: ignore[attr-defined]
            expected_endpoint,
            {
                "order_by": "period",
                "period_from": "2024-01-01T00:00:00+00:00",
                "period_to": "2024-01-31T00:00:00+00:00",
                "page_size": 10000,
                "group_by": "week",
            },
        )
        assert len(result) == 1



"""Tariff models."""

from datetime import datetime

from pydantic import BaseModel, Field


class HistoricalCharge(BaseModel):
    """Historical tariff charge (unit rate or standing charge).

    Represents a price that was valid during a specific time period.
    """

    value_exc_vat: float = Field(description="Price excluding VAT")
    value_inc_vat: float = Field(description="Price including VAT")
    valid_from: datetime = Field(description="Start of validity period")
    valid_to: datetime | None = Field(None, description="End of validity period")
    payment_method: str | None = Field(
        None,
        description="Payment method this charge applies to",
    )


class StandardElectricityTariff(BaseModel):
    """Single register electricity tariff.

    Standard tariffs have one rate that applies all day.
    """

    code: str = Field(description="Tariff code (e.g., E-1R-AGILE-24-04-03-A)")
    standing_charge_exc_vat: float = Field(description="Daily standing charge exc VAT")
    standing_charge_inc_vat: float = Field(description="Daily standing charge inc VAT")
    online_discount_exc_vat: float = Field(description="Online discount exc VAT")
    online_discount_inc_vat: float = Field(description="Online discount inc VAT")
    dual_fuel_discount_exc_vat: float = Field(
        description="Dual fuel discount exc VAT",
    )
    dual_fuel_discount_inc_vat: float = Field(
        description="Dual fuel discount inc VAT",
    )
    exit_fees_exc_vat: float = Field(description="Exit fees exc VAT")
    exit_fees_inc_vat: float = Field(description="Exit fees inc VAT")
    exit_fees_type: str = Field(
        description="Exit fee type (e.g., 'NONE', 'PER_FUEL')",
    )
    links: list[dict[str, str]] = Field(
        default_factory=list,
        description="HATEOAS links to rate endpoints",
    )


class Eco7ElectricityTariff(BaseModel):
    """Dual register electricity tariff (Economy 7).

    Economy 7 tariffs have different rates for day and night usage.
    """

    code: str = Field(description="Tariff code")
    standing_charge_exc_vat: float = Field(description="Daily standing charge exc VAT")
    standing_charge_inc_vat: float = Field(description="Daily standing charge inc VAT")
    online_discount_exc_vat: float = Field(description="Online discount exc VAT")
    online_discount_inc_vat: float = Field(description="Online discount inc VAT")
    dual_fuel_discount_exc_vat: float = Field(
        description="Dual fuel discount exc VAT",
    )
    dual_fuel_discount_inc_vat: float = Field(
        description="Dual fuel discount inc VAT",
    )
    exit_fees_exc_vat: float = Field(description="Exit fees exc VAT")
    exit_fees_inc_vat: float = Field(description="Exit fees inc VAT")
    exit_fees_type: str = Field(description="Exit fee type")
    links: list[dict[str, str]] = Field(
        default_factory=list,
        description="HATEOAS links to rate endpoints",
    )
    day_unit_rate_exc_vat: float | None = Field(
        None,
        description="Day unit rate exc VAT",
    )
    day_unit_rate_inc_vat: float | None = Field(
        None,
        description="Day unit rate inc VAT",
    )
    night_unit_rate_exc_vat: float | None = Field(
        None,
        description="Night unit rate exc VAT",
    )
    night_unit_rate_inc_vat: float | None = Field(
        None,
        description="Night unit rate inc VAT",
    )


class GasTariff(BaseModel):
    """Gas tariff."""

    code: str = Field(description="Tariff code (e.g., G-1R-AGILE-24-04-03-A)")
    standing_charge_exc_vat: float = Field(description="Daily standing charge exc VAT")
    standing_charge_inc_vat: float = Field(description="Daily standing charge inc VAT")
    online_discount_exc_vat: float = Field(description="Online discount exc VAT")
    online_discount_inc_vat: float = Field(description="Online discount inc VAT")
    dual_fuel_discount_exc_vat: float = Field(
        description="Dual fuel discount exc VAT",
    )
    dual_fuel_discount_inc_vat: float = Field(
        description="Dual fuel discount inc VAT",
    )
    exit_fees_exc_vat: float = Field(description="Exit fees exc VAT")
    exit_fees_inc_vat: float = Field(description="Exit fees inc VAT")
    exit_fees_type: str = Field(description="Exit fee type")
    links: list[dict[str, str]] = Field(
        default_factory=list,
        description="HATEOAS links to rate endpoints",
    )
    unit_rate_exc_vat: float | None = Field(None, description="Unit rate exc VAT")
    unit_rate_inc_vat: float | None = Field(None, description="Unit rate inc VAT")



"""Tests for AsyncOctopusClient."""

from unittest.mock import patch

import pytest

from pytentacle import AsyncOctopusClient
from pytentacle.api.electricity import ElectricityAPI
from pytentacle.api.gas import GasAPI
from pytentacle.api.industry import IndustryAPI
from pytentacle.api.products import ProductsAPI
from pytentacle.core.async_base import AsyncBaseClient


class TestAsyncOctopusClientInit:
    """Tests for AsyncOctopusClient initialization."""

    def test_init_with_api_key(self) -> None:
        """Test initialization with API key."""
        client = AsyncOctopusClient(api_key="test_key")

        assert client._base_client.auth is not None
        assert client._base_client.auth.api_key == "test_key"
        assert isinstance(client._base_client, AsyncBaseClient)

    @patch.dict("os.environ", {}, clear=True)
    def test_init_without_api_key(self) -> None:
        """Test initialization without API key."""
        client = AsyncOctopusClient()

        assert client._base_client.auth is None

    def test_init_with_custom_base_url(self) -> None:
        """Test initialization with custom base URL."""
        client = AsyncOctopusClient(base_url="https://custom.url")

        assert client._base_client.base_url == "https://custom.url"

    def test_init_with_custom_timeout(self) -> None:
        """Test initialization with custom timeout."""
        client = AsyncOctopusClient(timeout=60.0)

        assert client._base_client.timeout == 60.0

    def test_api_modules_initialized(self) -> None:
        """Test that all API modules are initialized."""
        client = AsyncOctopusClient(api_key="test_key")

        assert isinstance(client.products, ProductsAPI)
        assert isinstance(client.electricity, ElectricityAPI)
        assert isinstance(client.gas, GasAPI)
        assert isinstance(client.industry, IndustryAPI)


class TestAsyncOctopusClientMethods:
    """Tests for AsyncOctopusClient methods."""

    @pytest.mark.asyncio
    async def test_close(self) -> None:
        """Test closing the client."""
        client = AsyncOctopusClient(api_key="test_key")

        # Trigger client creation
        _ = client._base_client.client

        await client.close()
        assert client._base_client._client is None

    @pytest.mark.asyncio
    async def test_context_manager(self) -> None:
        """Test using client as async context manager."""
        async with AsyncOctopusClient(api_key="test_key") as client:
            assert client is not None
            assert isinstance(client, AsyncOctopusClient)

    def test_repr_authenticated(self) -> None:
        """Test string representation with authentication."""
        client = AsyncOctopusClient(api_key="test_key")

        assert repr(client) == "AsyncOctopusClient(authenticated)"

    @patch.dict("os.environ", {}, clear=True)
    def test_repr_unauthenticated(self) -> None:
        """Test string representation without authentication."""
        client = AsyncOctopusClient()

        assert repr(client) == "AsyncOctopusClient(unauthenticated)"



"""Meter point models."""

from pydantic import BaseModel, Field


class ElectricityMeterPoint(BaseModel):
    """Electricity meter point details.

    MPAN (Meter Point Administration Number) uniquely identifies
    an electricity supply point in the UK.
    """

    mpan: str = Field(description="Meter Point Administration Number")
    profile_class: int = Field(
        description="Profile class (1-8, indicates usage pattern)",
    )
    consumption_standard: int | None = Field(
        default=None,
        description="Standard consumption in kWh",
    )
    gsp: str = Field(
        description="Grid Supply Point group ID (e.g., '_A', '_B', '_H')",
    )


class GasMeterPoint(BaseModel):
    """Gas meter point details.

    MPRN (Meter Point Reference Number) uniquely identifies
    a gas supply point in the UK.
    """

    mprn: str = Field(description="Meter Point Reference Number")



"""Pytentacle - Modern Python client for the Octopus Energy API.

This library provides both synchronous and asynchronous clients for
interacting with the Octopus Energy API.

Example:
    ```python
    from pytentacle import OctopusClient

    client = OctopusClient(api_key="your_api_key")
    products = client.products.list()
    client.close()
    ```

    Or with async:
    ```python
    from pytentacle import AsyncOctopusClient

    async with AsyncOctopusClient(api_key="your_api_key") as client:
        products = await client.products.list()
    ```
"""

from pytentacle.async_client import AsyncOctopusClient
from pytentacle.client import OctopusClient
from pytentacle.exceptions import (
    APIError,
    AuthenticationError,
    NotFoundError,
    PytentacleError,
    RateLimitError,
    ServerError,
    ValidationError,
)
from pytentacle.models import (
    Consumption,
    Eco7ElectricityTariff,
    ElectricityMeterPoint,
    GasTariff,
    GridSupplyPoint,
    HistoricalCharge,
    Product,
    StandardElectricityTariff,
)

__version__ = "0.1.0"

__all__ = [
    "APIError",
    "AsyncOctopusClient",
    "AuthenticationError",
    "Consumption",
    "Eco7ElectricityTariff",
    "ElectricityMeterPoint",
    "GasTariff",
    "GridSupplyPoint",
    "HistoricalCharge",
    "NotFoundError",
    "OctopusClient",
    "Product",
    "PytentacleError",
    "RateLimitError",
    "ServerError",
    "StandardElectricityTariff",
    "ValidationError",
]



"""Nox configuration for running tests, linting, and documentation checks."""

import nox
from nox import Session

nox.options.sessions = ["tests", "lint", "mypy", "docs"]


@nox.session(python=["3.10", "3.11", "3.12", "3.13"])
def tests(session: Session) -> None:
    """Run the test suite.

    Args:
        session: The nox session object.
    """
    session.install(".[dev]")
    session.run("pytest", *session.posargs)


@nox.session(python="3.10")
def lint(session: Session) -> None:
    """Run the linter.

    Args:
        session: The nox session object.
    """
    session.install(".[dev]")
    session.run("ruff", "check", ".")
    session.run("ruff", "format", "--check", ".")


@nox.session(python="3.10")
def mypy(session: Session) -> None:
    """Run mypy.

    Args:
        session: The nox session object.
    """
    session.install(".[dev]")
    session.run("mypy", "src/pytentacle")


@nox.session(python="3.10")
def format_code(session: Session) -> None:
    """Format code with ruff.

    Args:
        session: The nox session object.
    """
    session.install(".[dev]")
    session.run("ruff", "format", ".")
    session.run("ruff", "check", "--fix", ".")


@nox.session(python="3.10")
def coverage(session: Session) -> None:
    """Generate coverage report.

    Args:
        session: The nox session object.
    """
    session.install(".[dev]")
    session.run(
        "pytest",
        "--cov=pytentacle",
        "--cov-report=html",
        "--cov-report=term",
        "--cov-report=lcov",
    )


@nox.session(python="3.10")
def docs(session: Session) -> None:
    """Check docstring quality and coverage.

    Args:
        session: The nox session object.
    """
    session.install(".[dev]")
    session.run("interrogate", "src/pytentacle", "--verbose")


@nox.session(python="3.10")
def docs_strict(session: Session) -> None:
    """Check docstring quality with strict argument checking.

    Args:
        session: The nox session object.
    """
    session.install(".[dev]")
    session.run("interrogate", "src/pytentacle", "--verbose")
    session.run("pydoclint", "src/pytentacle")



"""Synchronous Octopus Energy API client."""

import os

from dotenv import load_dotenv

from pytentacle.api.electricity import ElectricityAPI
from pytentacle.api.gas import GasAPI
from pytentacle.api.industry import IndustryAPI
from pytentacle.api.products import ProductsAPI
from pytentacle.core.base import BaseClient

# Load environment variables from .env file
load_dotenv()


class OctopusClient:
    """Synchronous client for the Octopus Energy API.

    This client provides access to all public Octopus Energy API endpoints.
    Most endpoints require authentication with an API key.

    Example:
        ```python
        from pytentacle import OctopusClient

        client = OctopusClient(api_key="sk_live_...")

        # List all products
        products = client.products.list()

        # Get specific product
        product = client.products.get("AGILE-24-04-03")

        # Get consumption data
        consumption = client.electricity.get_consumption(
            mpan="1234567890123",
            serial_number="12A3456789",
        )

        # Remember to close the client
        client.close()
        ```

    Or use as a context manager:
        ```python
        with OctopusClient(api_key="sk_live_...") as client:
            products = client.products.list()
        ```
    """

    def __init__(
        self,
        api_key: str | None = None,
        base_url: str | None = None,
        timeout: float = 30.0,
    ) -> None:
        """Initialize the Octopus Energy API client.

        Args:
            api_key: API key for authentication. If not provided, will attempt to
                read from OCTOPUS_API_KEY environment variable or .env file.
            base_url: Optional base URL (defaults to production API)
            timeout: Request timeout in seconds (default: 30.0)
        """
        # Use provided API key or fall back to environment variable
        if api_key is None:
            api_key = os.getenv("OCTOPUS_API_KEY")

        self._base_client = BaseClient(
            api_key=api_key,
            base_url=base_url,
            timeout=timeout,
        )

        # Initialize API modules
        self.products = ProductsAPI(self._base_client)
        self.electricity = ElectricityAPI(self._base_client)
        self.gas = GasAPI(self._base_client)
        self.industry = IndustryAPI(self._base_client)

    def close(self) -> None:
        """Close the HTTP client and release resources."""
        self._base_client.close()

    def __enter__(self) -> "OctopusClient":
        """Context manager entry."""
        return self

    def __exit__(self, *args: object) -> None:
        """Context manager exit."""
        self.close()

    def __repr__(self) -> str:
        """Return string representation."""
        auth_status = "authenticated" if self._base_client.auth else "unauthenticated"
        return f"OctopusClient({auth_status})"



"""API endpoint implementations."""

from pytentacle.api.electricity import ElectricityAPI
from pytentacle.api.gas import GasAPI
from pytentacle.api.industry import IndustryAPI
from pytentacle.api.products import ProductsAPI

__all__ = ["ElectricityAPI", "GasAPI", "IndustryAPI", "ProductsAPI"]



"""Consumption data models."""

from datetime import datetime

from pydantic import BaseModel, Field


class Consumption(BaseModel):
    """Energy consumption data for a time period.

    Half-hourly data is only available for smart meters.
    - Electricity consumption is in kWh
    - Gas consumption units vary: kWh for SMETS1, m³ for SMETS2
    """

    consumption: float = Field(description="Energy consumption amount")
    interval_start: datetime = Field(
        description="Start of consumption interval (inclusive)",
    )
    interval_end: datetime = Field(
        description="End of consumption interval (exclusive)",
    )
